/* gnome-region-panel-input.h
 * Copyright (C) 2011 Red Hat, Inc.
 *
 * Written by Matthias Clasen
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 */

#ifndef __GNOME_KEYBOARD_PROPERTY_INPUT_H
#define __GNOME_KEYBOARD_PROPERTY_INPUT_H

#include <gtk/gtk.h>

#include "cc-region-panel.h"

G_BEGIN_DECLS

void setup_input_tabs (GtkBuilder    *builder,
                       CcRegionPanel *self);

G_END_DECLS

#endif  /* __GNOME_KEYBOARD_PROPERTY_INPUT_H */
